import json
from abc import ABCMeta, abstractmethod
from functools import cache, cached_property
from threading import Lock
from typing import Optional

import requests

from ..driver import Driver
from ..state import Version
from ..utils import Some, unsome, url_version
from .base import Package


class Installer(metaclass=ABCMeta):
    def __init__(self, driver: Driver) -> None:
        self.driver = driver
        super().__init__()

    @abstractmethod
    def install(self, package: str) -> None:
        pass

    @abstractmethod
    def upgrade(self, package: str) -> None:
        pass

    def is_installed(self, package: str) -> bool:
        return self.installed_version(package) is not None

    @abstractmethod
    def installed_version(self, package: str) -> Optional[str]:
        pass

    @abstractmethod
    def latest_version(self, package: str) -> str:
        pass


class Brew(Installer):
    def install(self, package: str) -> None:
        self.driver.run("brew", "install", package)

    def upgrade(self, package: str) -> None:
        self.driver.run("brew", "upgrade", package)

    @cache
    def brew_info(self, package: str) -> dict:
        return json.loads(self.driver.run_output("brew", "info", "--json=v2", package))

    def installed_version(self, package: str) -> Optional[str]:
        info = self.brew_info(package)
        if info["casks"]:
            cask = info["casks"][0]
            return cask["installed"]
        if info["formulae"]:
            formula = info["formulae"][0]
            try:
                installed = formula["installed"][0]["version"]
            except IndexError:
                installed = None
            return installed
        raise ValueError(f"Unexpected output from brew: {info}")

    def api(self, path: str) -> dict:
        result = requests.get(f"https://formulae.brew.sh/api/{path}")
        result.raise_for_status()
        return result.json()

    CASK_PREFIX = "homebrew/cask/"

    @cache
    def latest_version(self, package: str) -> str:
        if package.startswith(self.CASK_PREFIX):
            # Cask
            # https://formulae.brew.sh/docs/api/#get-formula-metadata-for-a-cask-formula
            cask_package = package[len(self.CASK_PREFIX) :]
            cask = self.api(f"cask/{cask_package}.json")
            return cask["version"]
        elif "/" not in package:
            # Normal formula
            # https://formulae.brew.sh/docs/api/#get-formula-metadata-for-a-core-formula
            formula = self.api(f"formula/{package}.json")
            return self.formula_version(formula)
        else:
            # Non-core cask or formula?
            # https://github.com/Homebrew/discussions/discussions/3618
            info = self.brew_info(package)
            if info.get("formulae"):
                formula = info["formulae"][0]
                return self.formula_version(formula)
            elif info.get("casks"):
                cask = info["casks"][0]
                return cask["version"]
            else:
                raise ValueError(f"Unknown package: {package}")

    @staticmethod
    def formula_version(info: dict) -> str:
        version = info["versions"]["stable"]
        revision = info.get("revision")
        if revision:
            version += f"_{revision}"
        return version


class DNF(Installer):
    def install(self, package: str) -> None:
        self.driver.with_root(True).run("dnf", "install", "-y", package)

    def upgrade(self, package: str) -> None:
        self.driver.with_root(True).run("dnf", "upgrade", "-y", package)

    def installed_version(self, package: str) -> Optional[str]:
        return self.driver.run_(
            "rpm",
            "--query",
            "--queryformat",
            "%{VERSION}",
            "--whatprovides",
            package,
            check=False,
            silent=True,
            capture_output=True,
        ).output

    def latest_version(self, package: str) -> str:
        output = self.driver.run_output(
            "dnf",
            "--quiet",
            "repoquery",
            "--queryformat",
            "%{VERSION}",
            "--latest-limit",
            "1",
            "--arch",
            "x86_64,noarch",
            "--whatprovides",
            package,
        )
        if not output or "\n" in output:
            raise Exception(f"Cannot determine version for {package}.")
        return output


class Apt(Installer):
    def install(self, package: str) -> None:
        self.driver.with_root(True).run("apt", "install", "--yes", package)

    def upgrade(self, package: str) -> None:
        self.install(package)

    def latest_version(self, package: str) -> str:
        output = self.driver.run_output(
            "apt-cache", "show", "--quiet", "--no-all-versions", package
        ).strip()
        for line in output.splitlines():
            line = line.strip()
            if line.startswith("Version:"):
                return line.split(": ", 1)[-1]
        raise Exception(f"Cannot determine version for {package}.")

    def installed_version(self, package: str) -> Optional[str]:
        return self.driver.run_(
            "dpkg-query",
            "--showformat",
            "${Version}",
            "--show",
            package,
            check=False,
            silent=True,
            capture_output=True,
        ).output


def linux_installer(driver: Driver) -> Installer:
    if driver.executable_exists("dnf"):
        return DNF(driver)
    elif driver.executable_exists("apt"):
        return Apt(driver)
    else:
        raise NotImplementedError("Cannot find a package manager.")


INSTALLER_LOCK = Lock()


class SystemPackage(Package):
    def __init__(
        self,
        *,
        name: str,
        url: Optional[str] = None,
        auto_updates: bool = False,
        service: Some[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self._name = name
        self.url = url
        self.auto_updates = auto_updates
        self.services = unsome(service)

    @cached_property
    def installer(self):
        return self.driver.os.switch(
            linux=linux_installer(self.driver), macos=Brew(self.driver)
        )

    @property
    def name(self) -> str:
        return self._name

    def get_remote_version(self) -> str:
        if self.url:
            return url_version(self.url)
        if self.auto_updates:
            return "latest"
        return self.installer.latest_version(self.name)

    @property
    def local_version(self) -> Optional[str]:
        if self.url:
            try:
                return self.versions[self.name].version
            except KeyError:
                return None
        if self.auto_updates:
            return "latest" if self.installer.is_installed(self.name) else None
        return self.installer.installed_version(self.name)

    def postinstall_linux(self):
        if self.services:
            self.driver.with_root(True).run("systemctl", "daemon-reload")
            for service in self.services:
                self.driver.with_root(True).run("systemctl", "enable", service)

    def postinstall_macos(self):
        pass

    def install(self):
        with INSTALLER_LOCK:
            if self.url:
                self.installer.install(self.url)
                self.versions[self.name] = Version(version=self.get_remote_version())
            elif self.local_version:
                self.installer.upgrade(self.name)
            else:
                self.installer.install(self.name)
        self.driver.os.switch(
            linux=self.postinstall_linux, macos=self.postinstall_macos
        )()



import subprocess
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Callable, Iterable, Literal, Optional, cast

from .utils import TERMINAL_LOCK, T

LinkMethod = Literal["binary_wrapper"]


class OS(metaclass=ABCMeta):
    def __init__(self, driver: "Driver") -> None:
        self.driver = driver

    @abstractmethod
    def switch_(self, *, linux: Callable[["Linux"], T], macos: T) -> T:
        pass

    def switch(self, *, linux: T, macos: T) -> T:
        return self.switch_(linux=lambda _: linux, macos=macos)


class Linux(OS):
    def switch_(self, *, linux: Callable[["Linux"], T], macos: T) -> T:
        return linux(self)

    RELEASE_FILE = "/etc/os-release"

    @cached_property
    def distribution(self) -> str:
        for line in self.driver.read_file(Path(self.RELEASE_FILE)).splitlines():
            k, v = line.split("=", 1)
            if k == "ID":
                return v

        raise ValueError(f"Cannot find distribution ID in {self.RELEASE_FILE}.")


class MacOS(OS):
    def switch_(self, *, linux: Callable[["Linux"], T], macos: T) -> T:
        return macos


@dataclass
class RunResult:
    ok: bool
    output: Optional[str]


class Driver(metaclass=ABCMeta):
    def __init__(self, *, root: bool = False) -> None:
        self.root = root
        super().__init__()

    def deconstruct(self) -> dict:
        return {"root": self.root}

    def with_root(self, /, root: bool) -> "Driver":
        if self.root == root:
            return self
        kwargs = self.deconstruct() | {"root": root}
        return type(self)(**kwargs)

    @abstractmethod
    def run_(
        self,
        *args: str,
        check: bool = True,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        capture_output: bool = False,
        silent: bool = False,
    ) -> RunResult:
        pass

    def run(
        self,
        *args: str,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        silent: bool = False,
    ) -> None:
        self.run_(*args, input=input, silent=silent)

    def run_ok(self, *args: str) -> bool:
        return self.run_(*args, check=False).ok

    def run_output(self, *args: str, silent: bool = False) -> str:
        return cast(str, self.run_(*args, capture_output=True, silent=silent).output)

    def executable_exists(self, executable: str) -> bool:
        return self.run_ok("sh", "-c", f"command -v {executable}")

    def find_executable(self, *executables: str) -> str:
        for candidate in executables:
            if self.executable_exists(candidate):
                return candidate
        raise Exception(f"None of {', '.join(executables)} found in PATH.")

    def is_file(self, path: Path) -> bool:
        return self.run_ok("test", "-f", str(path))

    def is_executable(self, path: Path) -> bool:
        return self.run_ok("test", "-x", str(path))

    def is_dir(self, path: Path) -> bool:
        return self.run_ok("test", "-d", str(path))

    def home(self) -> Path:
        path = "~root" if self.root else "~"
        return Path(self.with_root(False).run_output("sh", "-c", f"eval echo {path}"))

    def local(self) -> Path:
        return self.home() / ".local"

    def makedirs(self, path: Path) -> None:
        self.run("mkdir", "-p", str(path))

    def rm(self, path: Path) -> None:
        self.run("rm", "-r", "-f", str(path))

    def make_executable(self, path: Path) -> None:
        self.run("chmod", "+x", str(path))

    def read_file(self, path: Path) -> str:
        return self.run_output("cat", str(path))

    def write_file(self, path: Path, content: str) -> None:
        self.run("cp", "/dev/stdin", str(path), input=content.encode())

    def link(
        self,
        source: Path,
        target: Path,
        *,
        method: Optional[LinkMethod] = None,
    ) -> None:
        self.makedirs(target.parent)
        self.rm(target)
        if method == "binary_wrapper":
            self.write_file(target, f'#!/bin/sh\nexec "{source}" "$@"')
            self.make_executable(target)
        else:
            self.run("ln", "-s", "-f", str(source), str(target))

    @cached_property
    def os(self) -> OS:
        driver = self.with_root(False)
        os_type = driver.run_output("uname")
        if os_type == "Linux":
            return Linux(driver)
        elif os_type == "Darwin":
            return MacOS(driver)
        else:
            raise ValueError(f"Unsupported OS type {os_type}.")


class SubprocessDriver(Driver, metaclass=ABCMeta):
    def prepare_command(self, args: Iterable[str]) -> list[str]:
        return list(args)

    def run_(
        self,
        *args: str,
        check: bool = True,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        capture_output: bool = False,
        silent: bool = False,
    ) -> RunResult:
        command = self.prepare_command(args)

        if capture_output:
            stdout = subprocess.PIPE
        elif not check:
            stdout = subprocess.DEVNULL
        else:
            stdout = None

        if not check or silent:
            stderr = subprocess.DEVNULL
        else:
            stderr = None

        result = subprocess.run(
            command, check=check, input=input, stdout=stdout, stderr=stderr
        )

        ok = result.returncode == 0
        output: Optional[str]
        if ok and capture_output:
            output = result.stdout.decode().strip()
        else:
            output = None
        return RunResult(ok=result.returncode == 0, output=output)


class LocalDriver(SubprocessDriver):
    def prepare_command(self, args: Iterable[str]) -> list[str]:
        if self.root:
            # If the sudo prompt is needed, avoid drawing a progress bar over it
            # with first prompting for a no-op command.
            with TERMINAL_LOCK:
                subprocess.run(["sudo", "true"], check=True)
            return super().prepare_command(["sudo", *args])
        else:
            return super().prepare_command(args)



import concurrent.futures
import re
import subprocess
from pathlib import Path
from threading import Lock
from typing import Any, Callable, Iterable, Iterator, Optional, TypeVar, Union

import requests
import tqdm  # type: ignore

T = TypeVar("T")
U = TypeVar("U")


Some = Optional[Union[T, list[T]]]


def unsome_(x: Some[T]) -> Optional[list[T]]:
    if x is None:
        return None
    if isinstance(x, list):
        return x
    return [x]


def unsome(x: Some[T]) -> list[T]:
    return unsome_(x) or []


TERMINAL_LOCK = Lock()


def run(*args: str) -> subprocess.CompletedProcess:
    return subprocess.run(args, check=True)


def run_ok(*args: str) -> bool:
    return (
        subprocess.run(
            args,
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ).returncode
        == 0
    )


def run_output(*args: str) -> str:
    return (
        subprocess.run(args, stdout=subprocess.PIPE, check=True).stdout.decode().strip()
    )


def flatten(items: Iterable[Iterable[T]]) -> list[T]:
    return [item for sublist in items for item in sublist]


def parallel_map_tqdm(action: Callable[[T], U], items: list[T]) -> list[U]:
    with tqdm.tqdm(total=len(items)) as progress:
        with concurrent.futures.ThreadPoolExecutor(20) as executor:

            def action_and_update(item: T) -> U:
                result = action(item)
                with TERMINAL_LOCK:
                    progress.update(1)
                return result

            return list(executor.map(action_and_update, items))


class Filters:
    @staticmethod
    def includes(substring: str) -> Callable[[str], bool]:
        return lambda x: substring in x.lower()

    @staticmethod
    def excludes(substring: str) -> Callable[[str], bool]:
        return lambda x: substring not in x

    @staticmethod
    def startswith(prefix: str) -> Callable[[str], bool]:
        return lambda x: x.startswith(prefix)

    @staticmethod
    def endswith(suffix: str) -> Callable[[str], bool]:
        return lambda x: x.endswith(suffix)

    @staticmethod
    def regex(regex: str) -> Callable[[str], bool]:
        regex_compiled = re.compile(regex)

        return lambda x: regex_compiled.match(x) is not None


def choose(candidates: list[T], filters: Iterator[Callable[[T], bool]]) -> T:
    if len(candidates) == 0:
        raise ValueError("No candidates to choose from.")
    while len(candidates) > 1:
        try:
            filter_fn = next(filters)
        except StopIteration:
            break

        new_candidates = list(filter(filter_fn, candidates))
        if new_candidates:
            candidates = new_candidates

    if len(candidates) == 1:
        return candidates[0]

    raise ValueError(f"Cannot choose between: {candidates}.")


def url_version(url: str) -> str:
    head_response = requests.head(url, allow_redirects=True)
    return head_response.headers["etag"]


def raise_(exception: BaseException) -> Any:
    raise exception


def transplant_path(dir_from: Path, dir_to: Path, path: Path) -> Path:
    return dir_to.joinpath(path.relative_to(dir_from))



from typing import Optional

from ..utils import run_output
from .destination import Destination


class Clone(Destination):
    def __init__(self, clone: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.repo = clone

    @property
    def name(self) -> str:
        return self.repo

    @property
    def directory_exists(self) -> bool:
        return self.driver.run_ok("test", "-d", str(self.destination))

    @property
    def git_args(self) -> list[str]:
        return ["git", "-C", str(self.destination)]

    def run_git(self, *args: str) -> None:
        self.driver.run(*self.git_args, *args)

    @property
    def local_version(self) -> Optional[str]:
        if not self.directory_exists:
            return None
        return self.driver.run_output(*self.git_args, "rev-parse", "HEAD")

    @property
    def remote(self):
        return f"https://github.com/{self.repo}.git"

    def get_remote_version(self) -> str:
        return run_output("git", "ls-remote", self.remote, "HEAD").split()[0]

    def install(self) -> None:
        self.driver.makedirs(self.destination.parent)
        if not self.directory_exists:
            self.driver.run("git", "clone", self.remote, str(self.destination))
        self.run_git("remote", "set-url", "origin", self.remote)
        self.run_git("fetch")
        default_branch = self.driver.run_output(
            *self.git_args, "rev-parse", "--abbrev-ref", "origin/HEAD"
        ).split("/")[1]
        self.run_git("switch", default_branch)
        self.run_git("reset", "--hard", f"origin/{default_branch}")



from abc import ABCMeta
from pathlib import Path

from .root import Root


class Destination(Root, metaclass=ABCMeta):
    destination: Path

    def __init__(self, destination: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.destination = self.driver.home() / destination



import tempfile
from abc import ABCMeta, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Optional, Union

from .manual import ManualPackage


class ArchivePackage(ManualPackage, metaclass=ABCMeta):
    def __init__(
        self,
        *,
        raw: Union[bool, str] = False,
        raw_executable: bool = False,
        strip: int = 0,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.raw = raw
        self.raw_executable = raw_executable
        self.strip = strip

    @abstractmethod
    def archive_url(self) -> str:
        pass

    def package_directory(self) -> Path:
        result = self.driver.local() / f"{self.name.replace('/', '--')}.app"
        self.driver.makedirs(result)
        return result

    @cached_property
    def tar(self) -> str:
        return self.driver.find_executable("gtar", "tar")

    def untar(self, source: Path, *extra: str) -> None:
        self.driver.run(
            self.tar,
            "-x",
            "--strip",
            str(self.strip),
            "-C",
            str(self.package_directory()),
            *extra,
            "-f",
            str(source),
        )

    def unzip(self, source: Path) -> None:
        if self.strip > 0:
            raise NotImplementedError("Strip is not supported for unzip.")
        self.driver.run(
            "unzip", "-o", "-qq", str(source), "-d", str(self.package_directory())
        )

    def extract(self, url: str, source: Path) -> None:
        if self.raw:
            if isinstance(self.raw, str):
                filename = self.raw
            else:
                filename = url.rsplit("/", 1)[-1]
            target = self.package_directory() / filename
            self.driver.run("cp", str(source), str(target))
            if self.raw_executable:
                self.driver.make_executable(target)
        elif url.endswith(".tar"):
            self.untar(source)
        elif url.endswith(".tar.gz") or url.endswith(".tgz"):
            self.untar(source, "-z")
        elif url.endswith(".tar.bz2"):
            self.untar(source, "-j")
        elif url.endswith(".tar.xz") or url.endswith(".txz"):
            self.untar(source, "-J")
        elif url.endswith(".zip"):
            self.unzip(source)
        else:
            raise ValueError(f"Unknown archive format: {url}")

    def binary_path(self, binary: str) -> Path:
        paths: list[list[str]] = [[], ["bin"]]
        for relative_path in paths:
            candidate = self.package_directory() / Path(*relative_path) / binary
            if self.driver.is_executable(candidate):
                return candidate
        raise ValueError(f"Cannot find {binary} in {self.package_directory()}.")

    def app_path(self, name: str) -> Path:
        candidate = (
            self.package_directory() / "share" / "applications" / f"{name}.desktop"
        )
        if self.driver.is_file(candidate):
            return candidate
        raise ValueError(
            f"Cannot find application '{name}' in {self.package_directory()}."
        )

    def icon_directory(self) -> Optional[Path]:
        candidate = self.package_directory() / "share" / "icons"
        if self.driver.is_dir(candidate):
            return candidate
        return None

    def font_path(self, name: str) -> Path:
        candidate = self.package_directory() / name
        if self.driver.is_file(candidate):
            return candidate
        raise ValueError(f"Cannot find font '{name}' in {self.package_directory()}.")

    def install(self):
        url = self.archive_url()
        with tempfile.NamedTemporaryFile() as archive_file:
            archive_path = Path(archive_file.name)
            self.driver.run(
                "curl",
                "--silent",
                "--show-error",
                "--location",
                "--output",
                archive_path,
                url,
            )
            self.extract(url, archive_path)
        super().install()



from abc import ABCMeta
from typing import Optional

from ..state import Version
from .base import Package


class ManualVersion(Package, metaclass=ABCMeta):
    @property
    def local_version(self) -> Optional[str]:
        try:
            return self.versions[self.name].version
        except KeyError:
            return None

    def install(self) -> None:
        super().install()
        self.versions[self.name] = Version(version=self.get_remote_version())



import json
from typing import Optional

from .pip_base import PipBasePackage


def pip_cmd(cmd: str, /, *args: str) -> list[str]:
    return ["python3", "-m", "pip", cmd, "--user", *args]


class PipPackage(PipBasePackage):
    def __init__(self, *, pip: str, **kwargs) -> None:
        self.package = pip
        super().__init__(**kwargs)

    def get_pip_versions(self) -> dict[str, str]:
        packages_json = self.driver.run_output(
            *pip_cmd("list", "--format", "json"), silent=True
        )
        packages = json.loads(packages_json)
        return {package["name"]: package["version"] for package in packages}

    @property
    def local_version(self) -> Optional[str]:
        return self.get_pip_versions().get(self.package)

    def install(self) -> None:
        cmd = "install" if self.local_version is None else "upgrade"
        self.driver.run(*pip_cmd(cmd, self.package))



from abc import ABCMeta, abstractmethod
from functools import cached_property
from typing import Optional

from ..driver import Driver
from ..state import DB, VERSIONS, Table, Version
from ..utils import Some, unsome_


class Package(metaclass=ABCMeta):
    os: Optional[list[str]]
    distribution: Optional[list[str]]

    def __init__(
        self,
        *,
        db: DB,
        driver: Driver,
        os: Some[str] = None,
        distribution: Some[str] = None,
    ) -> None:
        self.db = db
        self.driver_ = driver
        self.os = unsome_(os)
        self.distribution = unsome_(distribution)

    @property
    def driver(self) -> Driver:
        return self.driver_

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def get_remote_version(self) -> str:
        pass

    @cached_property
    def remote_version(self) -> str:
        return self.get_remote_version()

    @property
    @abstractmethod
    def local_version(self) -> Optional[str]:
        pass

    @property
    def is_installed(self) -> bool:
        return self.remote_version == self.local_version

    @abstractmethod
    def install(self) -> None:
        pass

    @property
    def applicable(self) -> bool:
        def check_os(name: str) -> bool:
            return self.os is None or name in self.os

        return self.driver.os.switch_(
            linux=lambda linux: check_os("linux")
            and (self.distribution is None or linux.distribution in self.distribution),
            macos=check_os("darwin"),
        )

    def ensure(self) -> bool:
        if not self.applicable:
            return False
        if self.is_installed:
            return False
        self.install()
        return True

    @cached_property
    def versions(self) -> Table[Version]:
        return VERSIONS(self.db)



import argparse

from .driver import LocalDriver
from .package import load_packages
from .state.base import DB, DB_PATH
from .utils import flatten, parallel_map_tqdm


def parser() -> argparse.ArgumentParser:
    result = argparse.ArgumentParser(description="Install the environment")
    result.add_argument("component", nargs="*", help="The component to install")
    result.add_argument(
        "--sequential", action="store_true", help="Install packages one at a time"
    )
    return result


def main():
    db = DB(DB_PATH)
    driver = LocalDriver()
    args = parser().parse_args()
    components: frozenset[str] = frozenset(args.component) | {"base"}
    packages = flatten(
        load_packages(component, db=db, driver=driver) for component in components
    )
    map_fn = map if args.sequential else parallel_map_tqdm
    results = map_fn(lambda p: p.name if p.ensure() else None, packages)
    installed = list(filter(None, results))
    if installed:
        print(
            f"{len(installed)} packages installed or updated: {', '   .join(installed)}"
        )
    else:
        print("Everything up to date.")



import configparser
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Optional

from ..utils import Some, transplant_path, unsome
from .manual_version import ManualVersion
from .root import Root


class ManualPackage(Root, ManualVersion, metaclass=ABCMeta):
    binaries: list[str]
    apps: list[str]
    fonts: list[str]

    def __init__(
        self,
        *,
        binary: Some[str] = None,
        binary_wrapper: bool = False,
        app: Some[str] = None,
        font: Some[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.binaries = unsome(binary)
        self.binary_wrapper = binary_wrapper
        self.apps = unsome(app)
        self.fonts = unsome(font)

    @property
    def local(self) -> Path:
        if self.root:
            return Path("/usr/local")
        else:
            return self.driver.local()

    @abstractmethod
    def binary_path(self, binary: str) -> Path:
        pass

    def install_binary(self, name: str) -> None:
        self.driver.link(
            self.binary_path(name),
            self.local / "bin" / name,
            method="binary_wrapper" if self.binary_wrapper else None,
        )

    @abstractmethod
    def app_path(self, name: str) -> Path:
        pass

    def icon_directory(self) -> Optional[Path]:
        return None

    def icon_name(self, app_path: Path) -> Optional[str]:
        config = self.driver.read_file(app_path)
        app = configparser.ConfigParser()
        app.read_string(config)
        return app["Desktop Entry"].get("Icon")

    def install_app(self, name: str) -> None:
        self.driver.os.switch(
            linux=self.install_app_linux, macos=self.install_app_macos
        )(name)

    def install_app_linux(self, name: str) -> None:
        path = self.app_path(name)
        target = self.local / "share" / "applications" / f"{name}.desktop"
        self.driver.link(path, target)
        icons_source = self.icon_directory()  # pylint:disable=assignment-from-none
        if icons_source:
            icons_target = self.local / "share" / "icons"
            icon = self.icon_name(path)
            if icon:
                icons = self.driver.run_output(
                    "find", str(icons_source), "-name", f"{icon}.*"
                ).splitlines()
                for icon_path in map(Path, icons):
                    target = transplant_path(icons_source, icons_target, icon_path)
                    self.driver.link(icon_path, target)

    def install_app_macos(self, name: str) -> None:
        # FIXME: copy to /Applications and/or ~/Applications; ensure names,
        # etc. are correct
        pass

    @abstractmethod
    def font_path(self, name: str) -> Path:
        pass

    def install_font(self, name: str) -> None:
        font_dir = self.driver.os.switch(
            linux=self.local / "share" / "fonts",
            macos=self.driver.home() / "Library" / "Fonts",
        )
        self.driver.makedirs(font_dir)
        source = self.font_path(name)
        target = font_dir / name
        self.driver.link(source, target)
        if self.driver.executable_exists("fc-cache"):
            self.driver.run("fc-cache", "-f", str(font_dir))

    def install(self) -> None:
        for binary in self.binaries:
            self.install_binary(binary)
        for app in self.apps:
            self.install_app(app)
        for font in self.fonts:
            self.install_font(font)
        super().install()



import os
import pwd
from functools import cached_property
from pathlib import Path
from typing import Optional

from .base import Package

SHELLS_FILE = Path("/etc/shells")


class Shell(Package):
    def __init__(self, shell: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.shell = Path(shell)

    @property
    def name(self) -> str:
        return "_shell"

    def get_remote_version(self) -> str:
        return str(self.shell)

    @property
    def local_version(self) -> Optional[str]:
        return pwd.getpwuid(os.getuid()).pw_shell

    @cached_property
    def all_shells(self) -> list[str]:
        return self.driver.read_file(SHELLS_FILE).splitlines()

    def install(self) -> None:
        if not self.driver.is_file(self.shell):
            raise ValueError(f"{self.shell} does not exist.")
        if not self.driver.is_executable(self.shell):
            raise ValueError(f"{self.shell} is not executable.")
        if str(self.shell) not in self.all_shells:
            self.driver.with_root(True).run(
                "tee", "-a", str(SHELLS_FILE), input=str(self.shell).encode()
            )
        self.driver.run("chsh", "-s", str(self.shell))



import json
from typing import Optional

from .pip_base import PipBasePackage

PIPX = ["python3", "-m", "pipx"]


class PipxPackage(PipBasePackage):
    def __init__(self, *, pipx: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.package = pipx

    def get_pipx_versions(self) -> dict[str, str]:
        pipx_list = json.loads(
            self.driver.run_output(*PIPX, "list", "--json", silent=True)
        )
        packages = (
            item["metadata"]["main_package"] for item in pipx_list["venvs"].values()
        )
        return {package["package"]: package["package_version"] for package in packages}

    @property
    def local_version(self) -> Optional[str]:
        return self.get_pipx_versions().get(self.package)

    def install(self) -> None:
        cmd = "install" if self.local_version is None else "upgrade"
        self.driver.run(*PIPX, cmd, self.package)
